###################################################################################################
#                              MIT Licence (C) 2023 Cubicpath@Github                              #
###################################################################################################
"""A simple way to get live Halo data straight from Halo Waypoint."""
# TODO: Add logging functionality
# TODO: Redo Events implementation
# TODO: Remove pyright comments when PySide6 signals and events are properly type hinted.
from ._version import __version__
from ._version import __version_info__

__author__ = 'Cubicpath@Github <cubicpath@pm.me>'
"""Author's information."""
